local M = {}

M.enabled = false
M.decreaseToInitial = false
M.size = 0.001
M.limit = nil

local ocean = nil
local initialWaterPosition = nil
local currentWaterPosition = nil
local oldMission = nil
local decrease = false

-- functions
local function getObject(className, preferredObjName)
    local objNames = scenetree.findClassObjects(className)
    if objNames and tableSize(objNames) > 0 then
        for _, name in pairs(objNames) do
            local obj = scenetree.findObject(name)
            if obj and (name == preferredObjName or not preferredObjName) then
                return obj
            end
        end
    else
        local obj = scenetree.findObject(className)
        if obj then return obj end
    end

    return nil
end

local function matrixToTable(mat)
    return {
        c0 = mat:getColumn(0),
        c1 = mat:getColumn(1),
        c2 = mat:getColumn(2),
        c3 = mat:getColumn(3)
    }
end

local function tableToMatrix(tbl)
    local mat = MatrixF(true)
    mat:setColumn(0, tbl.c0)
    mat:setColumn(1, tbl.c1)
    mat:setColumn(2, tbl.c2)
    mat:setColumn(3, tbl.c3)
    return mat
end

local function setup()
    M.enabled = false
    M.size = 0.001
    M.limit = nil

    ocean = nil
    initialWaterPosition = nil
    currentWaterPosition = nil
    decrease = false
end

local function getOcean()
    local waterPlane = getObject("WaterPlane")
    if waterPlane then
        initialWaterPosition = waterPlane.position
        currentWaterPosition = initialWaterPosition
    end

    return waterPlane
end

local function resetWaterLevel()
    if not ocean then return end
    ocean.position = initialWaterPosition
end

-- callbacks
local function onUpdate()
    local mission = getMissionFilename()
    if not oldMission or oldMission ~= mission then
        if mission == "" then return end
        oldMission = mission

        setup()
        ocean = getOcean()
    end

    if not M.enabled or not ocean then return end

    local c0 = ocean.position:getColumn(0)
    local c1 = ocean.position:getColumn(1)
    local c2 = ocean.position:getColumn(2)
    local c3 = ocean.position:getColumn(3)

    if M.decreaseToInitial then
        local initialZ = string.format("%.2f", currentWaterPosition:getColumn(3).z)
        local z = string.format("%.2f", c3.z)

        if z > initialZ then
            c3.z = c3.z - M.size
        elseif z < initialZ then
            c3.z = c3.z + M.size
        else
            M.enabled = false
        end
    else
        if M.decrease then
            local target = c3.z - M.size
            
            if M.limit then
                if target > M.limit then
                    c3.z = target
                end
            else
                c3.z = c3.z - M.size
            end
        else
            local target = c3.z + M.size
            
            if M.limit then
                if target < M.limit then
                    c3.z = target
                end
            else
                c3.z = c3.z + M.size
            end
        end
    end

    local target = {
        c0 = c0,
        c1 = c1,
        c2 = c2,
        c3 = c3
    }

    ocean.position = tableToMatrix(target) 
end

local function onExtensionUnloaded()
    resetWaterLevel()
    setup()
end

M.resetWaterLevel = resetWaterLevel

M.onExtensionUnloaded = onExtensionUnloaded
M.onUpdate = onUpdate

return M